import os

from ds.context import TestExecutorMixin
from ds.presets.docker_base import mixins
from ds.presets.docker_base import commands
from ds.presets.docker_base import DockerContext
from ds.executor import TestExecutor


def override_run_option_by_attr(mixin, attr, run_option=None, default=None):
    if issubclass(DockerContext, mixin):
        bases = (DockerContext, )
    else:
        bases = (mixin, DockerContext, )

    run_option = run_option or attr

    Context = type('Context', bases, {})
    context = Context()
    assert context.get_run_options().get(run_option) == default

    Context = type('Context', bases, {
        attr: None,
    })
    context = Context()
    assert context.get_run_options().get(run_option) == None

    value = 'TEST'
    Context = type('Context', bases, {
        attr: value,
    })
    context = Context()
    assert context.get_run_options().get(run_option) == value


def check_command_invoke(context_class, command_name, invoke_args, execute_args):
    class Context(TestExecutorMixin, context_class):
        pass
    context = context_class()
    context[command_name](invoke_args)
    context.executor.commit()
    assert context.executor.logs[0] == ([TestExecutor.CALL] + execute_args)


def check_command_exists(mixin, command_class):
    if issubclass(DockerContext, mixin):
        bases = (DockerContext, )
    else:
        bases = (mixin, DockerContext, )

    Context = type('Context', bases, {})
    context = Context()
    command = command_class(context)
    assert context[command.get_name()]


def test_user_mixin():
    assert not issubclass(DockerContext, mixins.UserMixin)
    override_run_option_by_attr(mixins.UserMixin, 'container_user',
                                'user', os.getuid())


def test_network_mixin():
    assert issubclass(DockerContext, mixins.NetworkMixin)
    override_run_option_by_attr(mixins.NetworkMixin, 'network', 'network')


def test_environment_mixin():
    assert issubclass(DockerContext, mixins.EnvironmentMixin)
    override_run_option_by_attr(mixins.NetworkMixin, 'container_environment',
                                'environment', default={})

    # class Context(DockerContext):
    #     pass
    # context = Context()
    # assert context.get_run_options().get('environment') == {}
    #
    # envs = {'a': 1}
    #
    # class Context(DockerContext):
    #     def get_envs(self):
    #         return envs
    # context = Context()
    # assert context.get_run_options().get('environment') == envs


def test_mounts_mixin():
    pass


def test_home_mounts_mixin():
    pass


def test_working_dir_mixin():
    assert not issubclass(DockerContext, mixins.WorkingDirMixin)
    override_run_option_by_attr(mixins.WorkingDirMixin, 'working_dir',
                                default='/app/')


def test_project_mounts_mixin():
    pass


def test_shell_mixin():
    assert issubclass(DockerContext, mixins.ShellMixin)
    # override_run_option_by_attr(mixins.ShellMixin, 'shell_entry',
    #                             default='/bin/bash')
    check_command_exists(mixins.ShellMixin, commands.Shell)
    check_command_exists(mixins.ShellMixin, commands.RootShell)

    class Context(mixins.ShellMixin, DockerContext):
        pass
    check_command_invoke(Context, 'shell', ('docker', 'exec', ''))



import os
from os.path import exists
from os.path import isdir
from os.path import join

from ds.path import get_additional_import


def get_modules(path):
    result = []
    for name in os.listdir(path):
        filename = join(path, name)
        if isdir(filename) and exists(join(filename, '__init__.py')):
            result.append(name)
            continue
        if not filename.endswith('.py') or name.startswith('__'):
            continue
        result.append(name.rsplit('.py', 1)[0])
    return sorted(set(result))


def find_contexts():
    result = []
    for path in get_additional_import():
        result += [(name, path) for name in get_modules(path)]
    return result



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import random
from os.path import join

from ds import context
from ds.environment import PwdEnvironment
from ds.utils.drop_empty import drop_empty


class QemuEnvironment(PwdEnvironment):
    env_filename = '.ds-qemu'


class QemuEnvironmentMixin(context.Context):
    def __init__(self, **options):
        self._qemu_enviroment = None
        super(QemuEnvironmentMixin, self).__init__(**options)

    @property
    def qemu_environment(self):
        if self._qemu_enviroment is None:
            self._qemu_enviroment = QemuEnvironment()
        return self._qemu_enviroment


class SocketOperateMixin(QemuEnvironmentMixin):
    def connect_to(self, address):
        protocol, address = address.split(':', 1)
        address, opts = address.split(',', 1)
        if protocol == 'unix':
            self.executor.append(('rlwrap', 'nc', '-U', address))
        elif protocol in ['tcp', 'telnet']:
            self.executor.append(['rlwrap', 'nc'] + address.split(':'))
        else:
            raise NotImplementedError

    def send_to(self, address, data):
        protocol, address = address.split(':', 1)
        address, opts = address.split(',', 1)
        data = data + '\n'
        if protocol == 'unix':
            self.executor.append(('nc', '-U', address), input=data)
            self.executor.commit()
        elif protocol in ['tcp', 'telnet']:
            self.executor.append(['nc'] + address.split(':'), input=data)
            self.executor.commit()
        else:
            raise NotImplementedError

    def format_address(self, protocol, path, **options):
        return ':'.join([
            protocol,
            ','.join(drop_empty(*[
                path,
                ','.join([key for key, value in options.items() if value]),
            ])),
        ])

    def format_unix(self, path, **options):
        return self.format_address('unix', path, **options)

    def format_tcp(self, path, **options):
        return self.format_address('tcp', path, **options)

    def format_telnet(self, path, **options):
        return self.format_address('telnet', path, **options)


class TelnetSerialMixin(SocketOperateMixin):
    @property
    def telnet_address(self):
        default = lambda: '127.0.0.1:{}'.format(random.randint(24000, 25000))
        return self.qemu_environment.get('telnet-address', default=default)

    @property
    def telnet(self):
        if not self.telnet_address:
            return
        nowait = not self.wait_telnet
        return self.format_telnet(self.telnet_address, server=True, nowait=nowait)


class TcpMonitorMixin(SocketOperateMixin):
    @property
    def monitor_address(self):
        default = lambda: '127.0.0.1:{}'.format(random.randint(24000, 25000))
        return self.qemu_environment.get('monitor-address', default=default)

    @property
    def monitor(self):
        if not self.monitor_address:
            return
        return self.format_tcp(self.monitor_address, server=True, nowait=True)


class UnixMonitorMixin(SocketOperateMixin):
    @property
    def monitor_path(self):
        default = join(self.project_root, '.qemu-monitor')
        return self.qemu_environment.get('monitor-path', default=default)

    @property
    def monitor(self):
        if not self.monitor_path:
            return
        return self.format_unix(self.monitor_path, server=True, nowait=True)


class TcpShellMixin(SocketOperateMixin):
    @property
    def shell_address(self):
        default = lambda: '127.0.0.1:{}'.format(random.randint(24000, 25000))
        return self.qemu_environment.get('shell-address', default=default)

    @property
    def shell(self):
        if not self.shell_address:
            return
        nowait = not self.wait_shell
        return self.format_telnet(self.shell_address, server=True, nowait=nowait)


class UnixShellMixin(SocketOperateMixin):
    @property
    def shell_path(self):
        default = join(self.project_root, '.qemu-shell')
        return self.qemu_environment.get('shell-path', default=default)

    @property
    def shell(self):
        if not self.shell_path:
            return
        nowait = not self.wait_shell
        return self.format_unix(self.shell_path, server=True, nowait=nowait)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from logging import getLogger
import random

from ds import context
from ds.presets.qemu_base import commands
from ds.presets.qemu_base import mixins


logger = getLogger(__name__)


class BaseQemuContext(mixins.SocketOperateMixin, context.Context):
    name = None

    enable_kvm = True

    memory = None
    smp = None

    soundhw = None

    display = None
    nographic = None

    cdrom = None
    hda = None
    hdb = None
    hdc = None
    hdd = None

    drive = ()
    net = ()
    device = ()
    usbdevice = 'tablet',
    serial = ()
    chardev = ()

    daemonize = False

    vnc = None

    monitor = None

    telnet = None
    wait_telnet = False

    shell = None
    wait_shell = False

    qemu_arch = 'x86_64'
    viewer_bin = 'xvncviewer'
    qemu_img = 'qemu-img'

    @property
    def qemu_bin(self):
        return 'qemu-system-{}'.format(self.qemu_arch)

    def get_run_options(self):
        return (
            ('-name', str(self.name)) if self.name else (),
            '-daemonize' if self.daemonize else (),
            '-enable-kvm' if self.enable_kvm else (),
            ('-m', str(self.memory)) if self.memory else (),
            ('-smp', str(self.smp)) if self.smp else (),
            ('-soundhw', self.soundhw) if self.soundhw else (),
            ('-display', self.display) if self.display else (),
            '-nographic' if self.nographic else (),
            [
                ('-serial', item)
                for item in self.serial
            ],
            [
                ('-net', item)
                for item in self.net
            ],
            [
                ('-drive', item)
                for item in self.drive
            ],
            [
                ('-usbdevice', item)
                for item in self.usbdevice
            ],
            [
                ('-device', item)
                for item in self.device
            ],
            [
                ('-chardev', item)
                for item in self.chardev
            ],
            ('-cdrom', self.cdrom) if self.cdrom else (),
            ('-hda', self.hda) if self.hda else (),
            ('-hdb', self.hdb) if self.hdb else (),
            ('-hdc', self.hdc) if self.hdc else (),
            ('-hdd', self.hdd) if self.hdd else (),
            ('-vnc', self.vnc) if self.vnc else (),
            ('-monitor', self.monitor) if self.monitor else (),
            ('-serial', self.shell) if self.shell else (),
            ('-serial', self.telnet) if self.telnet else (),
        )

    def get_commands(self):
        return super(BaseQemuContext, self).get_commands() + [
            commands.CreateDisk,
            commands.Start,
            commands.RunIso,
            commands.RunDisk,
            commands.RunIsoAndDisk,
            commands.Reset if self.monitor else None,
            commands.Stop if self.monitor else None,
            commands.Info if self.monitor else None,
            commands.AttachMonitor if self.monitor else None,
            commands.AttachShell if self.shell else None,
            commands.AttachTelnet if self.telnet else None,
            commands.Viewer if self.vnc else None,
            commands.Kill,
        ]


class BasicQemuContext(mixins.TcpMonitorMixin, mixins.TelnetSerialMixin,
                       BaseQemuContext):
    smp = None
    memory = 1024

    soundhw = 'ac97'

    net = 'nic', 'user',

    display = 'none'

    nographic = None
    daemonize = True

    vnc_host = '127.0.0.1'
    vnc_display = None

    def get_name(self):
        return self.project_name

    @property
    def name(self):
        return self.get_name()

    @property
    def vnc(self):
        display = self.vnc_display
        if display is None:
            default = random.randint(0, 300)
            display = self.qemu_environment.get('display', default=default)
        return '{host}:{display}'.format(host=self.vnc_host, display=display)



import pytest

from ds import utils
from ds.command import kebab_to_snake


class ComplexObject:
    a = 0
    b = 'test'
    c = list(range(3))
    d = None
complex_object = ComplexObject()


def test_flatten():
    case = ('a', 'b', ['c'], ('d', 'e', ('f', 'g'), 'h'))
    assert utils.flatten(case) == list('abcdefgh')


def test_interactive():
    try:
        utils.is_interactive()
    except Exception:
        pytest.fail('is_interactive')


def test_tty():
    try:
        utils.get_tty_size()
        utils.get_tty_width()
    except Exception:
        pytest.fail('tty')


def test_random_passwords():
    assert utils.generate_random_string(10, 'a') == ('a' * 10)
    assert len(utils.generate_random_string(10)) == 10


def test_pretty_printer(capsys):
    utils.pretty_print_object(complex_object)
    captured = capsys.readouterr()
    assert captured.out == 'a: 0\nb: \'test\'\nc: [0, 1, 2]\nd: None\n'


def test_drop_empty():
    drop = utils.drop_empty(None, 'a', {}, 'b', None, None, 'c', 0)
    assert drop == ['a', {}, 'b', 'c', 0]


@pytest.mark.parametrize('a,b', [
    ('', ''),
    ('TestCase', 'test-case'),
    ('1Test', '1-test'),
    ('Test1Case', 'test1-case'),
])
def test_drop_empty(a, b):
    assert kebab_to_snake(a) == b



from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import os
from logging import getLogger
from os.path import exists
from os.path import join
from shutil import copyfile

from six.moves import input

from ds import context
from ds.path import relative
from ds.path import get_additional_import
from ds.path import get_possible_imports
from ds.discover import find_contexts
from ds.command import Command
from ds.command import preset_base_command


logger = getLogger(__name__)


BASHRC = """
dscomplete={path}
[ -f $dscomplete ] && source $dscomplete
"""


class Context(context.Context):
    def get_commands(self):
        return super(Context, self).get_commands() + [
            InstallAutocomplete,
            ListContexts,
            OverridePreset,
        ]


class ListContexts(Command):
    short_help = 'Show all possible context modules'

    weight = preset_base_command()

    def invoke_with_args(self, args):
        print(' '.join([item[0] for item in find_contexts()]))


class InstallAutocomplete(Command):
    short_help = ''
    usage = '[<shell>]'

    weight = preset_base_command()

    def invoke_with_args(self, args):
        shell = args['<shell>'] or 'bash'
        script = relative('autocomplete', shell)
        if not exists(script):
            logger.error('Unknown shell: %s', shell)
            return
        print('Add to .bashrc:')
        print(BASHRC.format(path=script))


class OverridePreset(Command):
    short_help = 'Copy a preset to one of local directories'

    weight = preset_base_command()

    default_option = 'default'
    other_option = '(other)'

    def invoke_with_args(self, args):
        variants = [' '.join(item) for item in find_contexts()]
        preset = self.context.executor.\
            fzf(variants, prompt='Preset')
        if not preset:
            return
        src, src_path = preset.split(' ', 1)

        variants = [self.other_option, self.default_option]
        if src != self.default_option:
            variants.append(src)
        dst = self.context.executor.\
            fzf(variants, prompt='New context')
        if not dst:
            return
        if dst == self.other_option:
            try:
                dst = input('New context name: ')
            except EOFError:
                dst = None
        if not dst:
            return

        additional_import = get_additional_import()
        possible_imports = get_possible_imports()

        variants = additional_import + [
            path for path in possible_imports
            if path not in additional_import
        ]

        dst_path = self.context.executor.\
            fzf(variants, prompt='Install path', no_sort=True)
        if not dst_path:
            return

        if not exists(dst_path):
            os.makedirs(dst_path)

        src = join(src_path, src) + '.py'
        dst = join(dst_path, dst) + '.py'

        if exists(dst):
            confirm = 'File "{}" already exists. Override?'.format(dst)
            if not self.context.executor.yesno(confirm):
                return

        copyfile(src, dst)
        logger.info('New context copied to "%s"', dst)

        self.context.executor.edit_file(dst)



from datetime import datetime
from logging import getLogger
from os import makedirs
from os.path import join
from os.path import exists
from os.path import dirname

from ds import context
from ds.command import Command


logger = getLogger()


class Context(context.Context):
    def get_commands(self):
        return super(Context, self).get_commands() + [
            ('n', New),
            New,
            ('v', View),
            View,
            ('l', View),
        ]

    @property
    def base_path(self):
        return join(self.project_root, 'notes')

    @property
    def new_filename(self):
        now = datetime.now()
        filename = '{}.txt'.format(now.strftime('%H-%M-%S-%f'))
        return join(self.base_path, now.strftime('%Y%m%d'), filename)


class New(Command):
    def invoke_with_args(self, args):
        filename = self.context.new_filename
        path = dirname(filename)
        if not exists(path):
            makedirs(path)
        logger.info('Edit %s', filename)
        self.context.executor.edit_file(filename)


class View(Command):
    def invoke_with_args(self, args):
        self.context.executor.append((
            'ranger',
            '--cmd=flat 1',
            self.context.base_path,
        ))



from __future__ import unicode_literals
import os
from collections import namedtuple
from logging import getLogger
from os import execvp
from subprocess import PIPE
from subprocess import Popen

from ds.utils import flatten
from ds.utils import drop_empty


logger = getLogger(__name__)

ExecResult = namedtuple('ExecResult', ('code', 'stdout', 'stderr'))


class BaseExecutor(object):
    def append(self, args, **opts):
        raise NotImplementedError

    def commit(self, replace=False):
        raise NotImplementedError


class ExecutorShortcuts(BaseExecutor):
    def fzf(self, choices, multi=False, prompt=None, no_sort=False):
        input_ = '\n'.join(choices)
        args = ['fzf']
        if multi:
            args.append('--multi')
        if prompt:
            args.append('--prompt=' + prompt.rstrip() + ' ')
        if no_sort:
            args.append('--no-sort')
        self.append(args, input=input_, skip_stderr=True)
        result = self.commit().stdout.strip().split('\n')
        if not multi:
            return result[0]
        return result

    def yesno(self, prompt):
        return self.fzf(['no', 'yes'], prompt=prompt) == 'yes'

    def edit_file(self, filename):
        editor = os.environ.get('EDITOR')
        if not editor:
            logger.error('$EDITOR is not defined')
            return
        self.append([editor, filename])
        self.commit(replace=True)


class Executor(ExecutorShortcuts, BaseExecutor):
    def __init__(self):
        super(Executor, self).__init__()
        self._queue = []

    @property
    def is_empty_queue(self):
        return len(self._queue) == 0

    def append(self, args, **opts):
        args = drop_empty(*flatten(args))
        if not args:
            return
        self._queue.append((args, opts))

    def _call(self, args, **opts):
        logger.debug('Call with %s', args)

        skip_all = opts.get('skip_all', False)
        skip_stdout = opts.get('skip_stdout', skip_all)
        skip_stdin = opts.get('skip_stdin', skip_all)
        skip_stderr = opts.get('skip_stderr', skip_all)
        shell = opts.get('shell', False)
        input_ = opts.get('input', None)

        popen_kwargs = dict(
            stdin=None if skip_stdin else PIPE,
            stdout=None if skip_stdout else PIPE,
            stderr=None if skip_stderr else PIPE,
            shell=shell,
        )
        process = Popen(args, **popen_kwargs)
        stdout, stderr = process.communicate(input_)
        code = process.poll()
        if code:
            logger.info('Code: %s, stdout: %s, stderr: %s',
                        code, repr(stdout), repr(stderr))
        return ExecResult(code, stdout, stderr)

    def _replace(self, args, **opts):
        logger.debug('Replace with %s', args)
        execvp(args[0], args[:])

    def commit(self, replace=False):
        queue = self._queue
        self._queue = []
        for is_last, (item, opts) in iter_with_last(queue):
            if is_last and replace and not opts:  # TODO: opts
                self._replace(item, **opts)
                return
            value = self._call(item, **opts)
            if is_last:
                return value


class TestExecutor(Executor):
    CALL = 'call'
    REPLACE = 'replace'

    def __init__(self):
        super(TestExecutor, self).__init__()
        self._log = []

    @property
    def execute_log(self):
        return self._log

    def _call(self, args, **opts):
        self._log.append((self.CALL, args, opts))

    def _replace(self, args, **opts):
        self._log.append((self.REPLACE, args, opts))


def iter_with_last(items):
    for item in items[:-1]:
        yield False, item
    for item in items[-1:]:
        yield True, item



from __future__ import unicode_literals

from ds.presets.docker_base import BuildContext
from ds.presets.docker_base import DefaultNaming
from ds.presets.docker_base import UserMixin
from ds.presets.docker_base import HomeMountsMixin
from ds.presets.docker_base import ProjectMountMixin


class ContainerBuildContext(ProjectMountMixin, UserMixin, HomeMountsMixin,
                            DefaultNaming, BuildContext):
    pass



from .drop_empty import drop_empty
from .flatten import flatten
from .format_columns import format_columns
from .is_interactive import is_interactive
from .pretty_printer import pretty_print_object
from .term import get_tty_size
from .term import get_tty_width
from .cd import cd
from .crypto import generate_random_string
from .numerator import numerator



from __future__ import unicode_literals
from collections import OrderedDict
from os import getcwd
from os.path import abspath
from os.path import basename
from os.path import dirname
from os.path import exists
from os.path import expanduser
from os.path import join
from os.path import realpath

from cachetools import cached


HIDDEN_PREFIX = '.ds'


def clean_path(path):
    return abspath(realpath(path))


def relative(*parts, **kwargs):
    return clean_path(join(dirname(kwargs.pop('target', __file__)), *parts))


@cached({})
def pwd():
    return getcwd()


@cached({})
def home():
    return join(expanduser('~'), HIDDEN_PREFIX)


@cached({})
def walk_top(path=None):
    path = clean_path(path or pwd())
    result = []
    while True:
        if path in result:
            continue
        result.append(path)
        path = dirname(path)
        if result[-1] == path:
            break
    return result


@cached({})
def find_for(target):
    for path in walk_top():
        if not exists(join(path, target)):
            continue
        if path == expanduser('~'):
            continue
        return path
    return pwd()


@cached({})
def get_project_name():
    return basename(find_for(HIDDEN_PREFIX))


@cached({})
def get_project_root():
    return find_for(HIDDEN_PREFIX)


@cached({})
def get_possible_imports():
    result = OrderedDict()
    for item in walk_top():
        result[join(item, HIDDEN_PREFIX)] = None
    result[home()] = None
    result[relative('presets')] = None
    return list(result.keys())


@cached({})
def get_additional_import():
    return filter(exists, get_possible_imports())



# import os
#
# from os.path import join, expanduser
#
# from ds.presets.docker_base.mixins import UserMixin
# from ds.presets.docker_base.mixins import HomeMountsMixin
#
#
# def test_user():
#     assert UserMixin().get_run_options().get('user') == os.getuid()
#
#     class Context(UserMixin):
#         container_user = 999
#
#     assert Context().get_run_options().get('user') == Context.container_user
#
#
# def test_home_mounts():
#     class Context(HomeMountsMixin):
#         container_home = 'a', 'b',
#         home_mounts = [
#             '.',
#             '..',
#         ]
#
#     context = Context()
#
#     mounts = {
#         item['Target']: item['Source']
#         for item in context.get_run_options().get('mounts')
#     }
#
#     home = expanduser('~')
#
#     for dest in context.container_home:
#         for src in context.home_mounts:
#             full_dst = join(dest, src)
#             assert full_dst in mounts
#
#             full_src = join(home, src)
#             assert mounts[full_dst] == full_src



# from ds.environment import get_environment
#
#
# context = 'container_simple'
#
#
# def test_switch_context(shell):
#     shell.call_ds('switch-context', context)
#     assert context == get_environment().get('context')
#
#
# def test_bad_cmd(shell):
#     shell.call_ds('switch-context', context)
#     result = shell.call_docker_ds('start', 'none')
#     assert result.code == 1
#
#
# def test_echo(shell):
#     shell.call_ds('switch-context', context)
#     result = shell.call_docker_ds('start', '/bin/bash', '-c', 'echo ok')
#     assert result.code == 0
#     assert result.stdout.strip() == 'ok'



from __future__ import unicode_literals
from logging import getLogger

import docker
import docker.errors

from ds import context
from . import naming


logger = getLogger()


class BaseDockerContext(naming.ContainerNaming, context.Context):
    def __init__(self):
        super(BaseDockerContext, self).__init__()
        self._client = None

    @property
    def client(self):
        if self._client is None:
            self._client = docker.from_env()
        return self._client

    @property
    def container(self):
        if not self.container_name:
            return
        try:
            return self.client.containers.get(self.container_name)
        except docker.errors.NotFound:
            pass

    def get_run_options(self, **options):
        """
        https://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.ContainerCollection.run
        """
        result = dict(
            detach=False,
            auto_remove=True,
            stdin_open=True,
            tty=True,
        )
        result.update(options)
        return result

    def filter_commands(self, commands):
        result = []
        for command in commands:
            if command.container_name_required and not self.has_container_name:
                logger.debug('Filter command %s', command)
                continue
            if command.image_name_required and not self.has_image_name:
                logger.debug('Filter command %s', command)
                continue
            result.append(command)
        return super(BaseDockerContext, self).filter_commands(result)

    def calc_cpu_to_options(self, cpu=1.0):
        period = int(cpu * 100000)
        quota = int(cpu * 50000)
        return {
            'cpu_period': period,
            'cpu_quota': quota,
        }



from logging import getLogger
from os import listdir
from os.path import join
from os.path import exists

from ds import context
from ds.path import relative
from ds.command import Command


logger = getLogger()


class Context(context.Context):
    files = 'incoming', 'backlog', 'doing', 'done'
    commit_with_empty_message = False

    def get_commands(self):
        return super(Context, self).get_commands() + [
            ('i', Init),
            ('e', Edit),
            ('s', Status),
            ('d', Diff),
            ('c', Commit),
        ]

    @property
    def templates_path(self):
        return self.get_templates_path()

    def get_templates_path(self):
        return relative('presets/todo/templates')

    @property
    def templates(self):
        return listdir(self.templates_path)

    @property
    def commit_message(self):
        return '' if self.commit_with_empty_message else '.'


class Init(Command):
    short_help = 'Init files and git'

    def invoke_with_args(self, args):
        src_path = self.context.templates_path
        files = self.context.templates

        some_exists = False
        for name in ['.git'] + files:
            if not exists(name):
                continue
            logger.error('%s already exists!', name)
            some_exists = True

        if some_exists:
            return

        self.context.executor.append(('git', 'init'))

        for filename in files:
            self.context.executor.append(('cp', join(src_path, filename), filename))
            self.context.executor.append(('git', 'add', filename))

        self.context.executor.append(('git', 'commit', '-m', 'init'))


class Edit(Command):
    short_help = 'Open vim with session'

    def invoke_with_args(self, args):
        self.context.executor.append(('vim', '-c', 'source .session.vim'),
                                     skip_all=True)
        self.context.c()


class Commit(Command):
    short_help = 'Commit all'

    def invoke_with_args(self, args):
        files = self.context.templates

        for filename in files:
            self.context.executor.append(('git', 'add', filename))
        self.context.executor.append(('git', 'commit',
                                      '--allow-empty-message',
                                      '-m', self.context.commit_message))


class Status(Command):
    short_help = 'git status'

    def invoke_with_args(self, args):
        self.context.executor.append(('git', 'status'))


class Diff(Command):
    short_help = 'git diff'

    def invoke_with_args(self, args):
        self.context.executor.append(('git', 'diff'))



from __future__ import unicode_literals
import sys
import importlib
import logging
from logging import basicConfig
from logging import getLogger
from logging import DEBUG
from logging import ERROR
from logging import INFO
from logging import WARNING

from docopt import docopt

from ds import __version__ as version
from ds.environment import get_environment
from ds.path import get_additional_import
from ds.utils import format_columns
from ds import errors


logger = getLogger(__name__)

PRE_USAGE = """usage: ds [-v|-vv|-vvv] [--version] [<args>...]

Options:
 -v|-vv|-vvv   Verbosity level
"""

USAGE = """usage: ds [-v|-vv|-vvv] [--version] <command> [<args>...]

Options:
 -v|-vv|-vvv   Verbosity level 

{additional}

Commands:
{commands}
"""


def format_usage(context, show_hidden=False):
    commands = format_columns(*[
        [name, command.short_help]
        for name, command in context.commands.items()
        if not command.hidden or show_hidden
    ])

    additional = '\n\n'.join([
        '{title}:\n{content}'.format(
            title=item.title,
            content=item.render(),
        )
        for item in context.get_additional_summary()
    ])

    return USAGE.format(commands=commands or '', additional=additional)


def invoke_context(context_class):
    pre_usage_args = [arg for arg in sys.argv[1:]
                      if arg not in ('-h', '--help')]
    pre_usage_options = docopt(PRE_USAGE, version=version, options_first=True,
                               argv=pre_usage_args)

    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    verbose_level = pre_usage_options.get('-v')
    level = {
        1: WARNING,
        2: INFO,
        3: DEBUG,
    }.get(verbose_level, ERROR)
    basicConfig(level=level)

    context = context_class()

    usage_options = docopt(format_usage(context, verbose_level >= 2),
                           version=version, options_first=True)

    name = usage_options.get('<command>')
    options = usage_options.get('<args>')

    if name not in context.commands:
        logger.error(errors.COMMAND_NOT_FOUND_MESSAGE)
        sys.exit(errors.COMMAND_NOT_FOUND)

    context.check()

    context[name].invoke(command_line=options)
    context.executor.commit(replace=True)


def context_fallback(context_name):
    logger.info('Fallback to default context. It was %s', context_name)
    get_environment().set('context', None)


def main():
    basicConfig(level=INFO)
    context_name = get_environment().get('context', None)

    no_context = context_name is None
    if no_context:
        context_name = 'default'

    sys.path = get_additional_import() + sys.path

    try:
        context_module = importlib.import_module(context_name)
    except ImportError as error:
        if not no_context:
            logger.exception(error)
            context_fallback(context_name)
            sys.exit(errors.NO_CONTEXT_MODULE)
        from ds.presets import dsjk as context_module

    if not hasattr(context_module, 'Context'):
        logger.error(errors.NO_CONTEXT_CLASS_MESSAGE)
        context_fallback(context_name)
        sys.exit(errors.NO_CONTEXT_CLASS)

    context_class = getattr(context_module, 'Context')
    invoke_context(context_class)


if __name__ == '__main__':
    main()



from os import urandom

from string import printable


def _normalize_hex(value, base):
    return int(round(value * base / 256))


def _byte_to_int(value):
    return int(value.encode('hex'), 16)


def generate_random_string(length=32, alphabet=None):
    alphabet = alphabet or printable
    return ''.join([
        alphabet[_normalize_hex(_byte_to_int(item), len(alphabet))]
        for item in urandom(length)
    ])



from __future__ import unicode_literals
import sys

try:
    import docker
    import docker.errors
    from docker.types import Mount
    # TODO: check version >= 3.1.0
except ImportError:
    print('Install docker-py with `pip install docker`')
    sys.exit(1)

from .contexts import BaseDockerContext
from .contexts import DockerContext
from .contexts import ExternalContext
from .contexts import BuildContext
from .contexts import PullContext
from .naming import ContainerNaming
from .naming import ImageNaming
from .naming import DefaultNaming
from .mixins import UserMixin
from .mixins import HomeMountsMixin
from .mixins import ProjectMountMixin



from ds.context import BaseContext
from ds.presets.docker_base.commands import Exec


class IPythonMixin(BaseContext):
    def get_commands(self):
        return super(IPythonMixin, self).get_commands() + [
            Ipython,
        ]


class PythonMixin(BaseContext):
    def get_commands(self):
        return super(PythonMixin, self).get_commands() + [
            Python,
        ]


class Python(Exec):
    def get_command(self):
        return 'python',


class Ipython(Exec):
    def get_command(self):
        return 'ipython',
